%  PhyloPattern
%  Copyright (C) 2005-2009 Philippe Gouret
%  This library is free software; you can redistribute it and/or
%  modify it under the terms of the GNU General Public
%  License as published by the Free Software Foundation; either
%  version 2 of the License, or (at your option) any later version.
%  This library is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%  General Public License for more details.
% 
%  You should have received a copy of the GNU General Public
%  License along with this library; if not, write to the
%  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
%  Boston, MA  02111-1307, USA. The text ol license can be also found 
%  at http://www.gnu.org/copyleft/gpl.html
%

%###############################################################################
%
%			PhyloPattern a Prolog tool for phylogeny tree analysis
%
%					Philippe Gouret 2006-2009
%
%###############################################################################


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      User predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% annotate tree => modify or add new tags values to all nodes of the tree
% +tree:				tree to be annotated
% +Annotation_function:	function to be used on each node to annotate, syntax: functor(list_of_additional_parameters)
% -Final_Tree:			the annotated tree
% +Order:				 'preorder' or 'postorder' atom, to choose the kind of traversal 
annotate_tree(Tree, Annotation_function, Final_Tree, preorder) :-
	preorder_annotate(Tree, Annotation_function, Final_Tree).
annotate_tree(Tree, Annotation_function, Final_Tree, postorder) :-
	postorder_annotate(Tree, Annotation_function, Annotated_Tree),
	unfreeze_tree(Annotated_Tree, Final_Tree).
	
% tree fullfills pattern and constraints
% +Tree:		tree to search on
% +Pattern:		tree pattern
% +Constraints:	list of predicates to be verified, syntax of one predicate: freeze(X, functor(X,.......))
% -Result:		result of Pattern applied on Tree, can be a subtree of a list of leafs
pattern_matching(Tree, Pattern, Constraints, Result) :-
	match(Result, Tree, Pattern),
	unfreeze_all(Constraints).
	
% gives all matching subtrees between two trees
% +Tree1: 				first tree
% +Tree2: 				second tree
% -MatchingSubtrees: 	list of Tree1's subtrees matching with at least of subtree of Tree2
% +EvenLeafs: 'true' gives also common leafs between the two trees, 'false' gives only internal nodes
matching_trees(Tree1, Tree2, MatchingSubtrees, EvenLeafs) :-
	bagof_always_successfull(Subtree, matching_subtree(Tree1, Tree2, Subtree, EvenLeafs, _), Subtrees),
	biggest_trees(Subtrees, [], MatchingSubtrees, self_tree_extractor).
	
% transform nodes with "bad" bootstraps values to rakes; and keep only "name" tag for nodes
% used before tree comparing, to match wired structures of trees
% +Tree: 	input tree
% -Result:	transformed tree
wired_and_broken_tree(Tree, Result) :-	break_bad_nodes(Tree, Result, _).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      Tree Annotation tools
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% get all tags with given functor
get_all_tags([], _, []).
get_all_tags([TagTerm|Tags], Tag, [TagTerm|SelectedTags]) :-
	TagTerm =.. [Tag, Value],
	get_all_tags(Tags, Tag, SelectedTags), !.
get_all_tags([TagTerm|Tags], Tag, SelectedTags) :-
	TagTerm =.. [TagX, _],
	diff(Tag, TagX),
	get_all_tags(Tags, Tag, SelectedTags), !.

% get tag value in node (has to be mono valued)
get_tag(Tags, Tag, Value) :-
	TagTerm =.. [Tag, Value],
	element(TagTerm, Tags),
	!.
	
% eliminate tag value from node 
eliminate_tag([], _, []) :- !.
eliminate_tag([TagTerm|Tags], Tag, NewTags) :-
	TagTerm =.. [Tag, _],
	eliminate_tag(Tags, Tag, NewTags), !.
eliminate_tag([TagTerm|Tags], Tag, [TagTerm|NewTags]) :-
	eliminate_tag(Tags, Tag, NewTags).
	
% eliminate freeze() terms from node
eliminate_freezes(_, Tags, NewTags, _) :-
	eliminate_tag(Tags, freeze, NewTags).
	
% set tag value in node (replace previous value (doesn't add the tag))
set_tag(Tags, Tag, Value, NewTags) :-
	TagTerm =.. [Tag, X],
	TagV =.. [Tag, Value],
	without(TagTerm, Tags, TagsM),
	concat(TagsM, [TagV], NewTags), !.

% call for element
call_for_element(List, OrderCall, Func, Result) :-
	element(X, List),
	Call =.. [OrderCall, X, Func, Result],
	Call.

% pre order annotation
preorder_annotate([[], Tags], Annotation_function, [[], NewTags]) :-
	Annotation_function =.. [F, Parameters],
	Func =.. [F, [], Tags, NewTags, Parameters],
	Func.
preorder_annotate([Children, Tags], Annotation_function, [NewChildren, NewTags]) :-
	diff(Children, []),
	Annotation_function =.. [F, Parameters],
	Func =.. [F, Children, Tags, NewTags, Parameters],
	Func,
	bagof_always_successfull(
		NewChild,
		call_for_element(Children, preorder_annotate, Annotation_function, NewChild),
		NewChildren).

% post order annotation
postorder_annotate([[], Tags], Annotation_function, [[], NewTags]) :-
	Annotation_function =.. [F, Parameters],
	Func =.. [F, [], Tags, NewTags, Parameters],
	Func.
postorder_annotate([Children, Tags], Annotation_function, [NewChildren, NewTags]) :-
	diff(Children, []),
	bagof_always_successfull(
		NewChild,
		call_for_element(Children, postorder_annotate, Annotation_function, NewChild),
		NewChildren),
	Annotation_function =.. [F, Parameters],
	Func =.. [F, NewChildren, Tags, NewTags, Parameters],
	Func.
	
% unfreeze tree
unfreeze_tree(Tree, Final_Tree) :-
	Tree = [Children, Tags],
	get_all_tags(Tags, freeze, FreezeTags),
	unfreeze_all(FreezeTags),
	preorder_annotate(Tree, eliminate_freezes([]), Final_Tree).
	
% unfreeze constraints
unfreeze_all([]) :- !.
unfreeze_all(FreezeTags) :-
	unfreeze(FreezeTags, [], NewFreezeTags),
	diff(FreezeTags, NewFreezeTags),
	unfreeze_all(NewFreezeTags).

% one unfreeze pass
unfreeze([], Left, Left).
unfreeze([freeze([X, FX])|Freezes], CurrentLeft, Left) :-
	nonvar(X),
	FX,
	!,
	unfreeze(Freezes, CurrentLeft, Left).
unfreeze([Freeze|Freezes], CurrentLeft, Left) :-
	concat(CurrentLeft, [Freeze], NewCurrentLeft),
	unfreeze(Freezes, NewCurrentLeft, Left).
		

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% annotate by calculating topological distance to a specified node (binary tree)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% two nodes are linked on the tree if
arc(A, B) :- is(B, A+1).
arc(A, B) :- is(B, A-1).

% dist is 0 if we are on the specified node, else set constraints between nodes
topological_distance([], Tags, NewTags, [TargetName]) :-
	get_tag(Tags, name, TargetName),
	set_tag(Tags, dist, D, NewTags),
	D = 0, !.
topological_distance([], Tags, NewTags, _) :-
	set_tag(Tags, dist, _, NewTags).
topological_distance([[_, Tags1], [_, Tags2]], Tags, NewTags, [TargetName]) :-
	get_tag(Tags, name, Name),
	diff(Name, TargetName),
	get_tag(Tags1, dist, D1),
	get_tag(Tags2, dist, D2),
	concat(Tags, [freeze([D, arc(D, D1)]),	freeze([D, arc(D, D2)]),
		freeze([D1, arc(D1, D)]), freeze([D2, arc(D2, D)])], NewTags00),
	set_tag(NewTags00, dist, D, NewTags0),
	get_all_tags(Tags1, freeze, C1),
	get_all_tags(Tags2, freeze, C2),
	concat(C1, C2, All),
	concat(NewTags0, All, NewTags).

% ### VERY VERY DIRTY: problem with rounding if A = B + C then B ~ A - C don't give the same results
float_is(A, B) :- var(A), is(A, B).
float_is(A, B) :- nonvar(A), is(A1, B), is(Delta, A-A1), Delta > -0.0000000001, Delta < 0.0000000001.
	
% two nodes are linked on the tree if
arc(N1, N2, A, B, L) :- float_is(B, A+L).
arc(N1, N2, A, B, L) :- float_is(B, A-L).

% dist is 0 if we are on the specified node, else set constraints between nodes
topological_distance2([], Tags, NewTags, [TargetName]) :-
	get_tag(Tags, id, TargetName),
	set_tag(Tags, dist, D, NewTags),
	D = 0.0, !.
topological_distance2([], Tags, NewTags, _) :-
	set_tag(Tags, dist, _, NewTags).
topological_distance2([[_, Tags1], [_, Tags2]], Tags, NewTags, [TargetName]) :-
	get_tag(Tags1, dist, D1),
	get_tag(Tags2, dist, D2),
	get_tag(Tags1, id, ID1),
	get_tag(Tags2, id, ID2),
	get_tag(Tags1, length, AL1),
	get_tag(Tags2, length, AL2),
	atom2number(AL1, L1),
	atom2number(AL2, L2),
	concat(Tags, [freeze([D, arc(ID1, ID2, D, D1, L1)]), freeze([D, arc(ID1, ID2, D, D2, L2)]),
	    freeze([D1, arc(ID1, ID2, D1, D, L1)]), freeze([D2, arc(ID1, ID2, D2, D, L2)])], NewTags00),
	set_tag(NewTags00, dist, D, NewTags0),
	(get_tag(Tags, id, TargetName) -> D = 0.0 ; D = D),
	get_all_tags(Tags1, freeze, C1),
	get_all_tags(Tags2, freeze, C2),
	concat(C1, C2, All),
	concat(NewTags0, All, NewTags).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%				annotate by giving all species under the node (binary tree)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% build on each node the list of all species in subtrees
all_species([], Tags, NewTags, _) :-
	get_tag(Tags, taxeid, S),
	set_tag(Tags, all_species, [S], NewTags), !.
all_species([[_, Tags1], [_, Tags2]], Tags, NewTags, _) :-
	get_tag(Tags1, all_species, S1),
	get_tag(Tags2, all_species, S2),
	concat(S1, S2, All),
	set_tag(Tags, all_species, All, NewTags).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%				pattern matching (binary tree)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% break bad nodes (bad bootstrap)
break_bad_nodes([[], Tags], [[], [name(Name)]], false) :- element(name(Name), Tags).
break_bad_nodes([[Child1, Child2], Tags], [[BrokenChild1, BrokenChild2], FreeTags], false) :-
	good_bootstraps(Tags),
	break_bad_nodes(Child1, BrokenChild1, _),
	break_bad_nodes(Child2, BrokenChild2, _),
	!.
break_bad_nodes([[Child1, Child2], Tags], &([BrokenChild1, BrokenChild2]), true) :-
	not(good_bootstraps(Tags)),
	break_bad_nodes(Child1, BrokenChild1, false),
	break_bad_nodes(Child2, BrokenChild2, false),
	!.
break_bad_nodes([[Child1, Child2], Tags], &(BrokenChildren), true) :-
	not(good_bootstraps(Tags)),
	break_bad_nodes(Child1, &(BrokenChild1), true),
	break_bad_nodes(Child2, BrokenChild2, false),
	concat(BrokenChild1, [BrokenChild2], BrokenChildren),
	!.
break_bad_nodes([[Child1, Child2], Tags], &(BrokenChildren), true) :-
	not(good_bootstraps(Tags)),
	break_bad_nodes(Child1, BrokenChild1, false),
	break_bad_nodes(Child2, &(BrokenChild2), true),
	concat([BrokenChild1], BrokenChild2, BrokenChildren),
	!.
break_bad_nodes([[Child1, Child2], Tags], &(BrokenChildren), true) :-
	not(good_bootstraps(Tags)),
	break_bad_nodes(Child1, &(BrokenChild1), true),
	break_bad_nodes(Child2, &(BrokenChild2), true),
	concat(BrokenChild1, BrokenChild2, BrokenChildren),
	!.


% leaf list
leaf_list([[], Tags], [[[], Tags]]).
leaf_list([[Child1, Child2], _], List) :-
	leaf_list(Child1, List1),
	leaf_list(Child2, List2),
	concat(List1, List2, List).
leaf_list(#(List), List).
leaf_list(&(Nodes), List) :- nodes_leafs(Nodes, [], List).
	
% nodes leafs
nodes_leafs([], Leafs, Leafs).
nodes_leafs([Node|Nodes], Leafs, Results) :-
	leaf_list(Node, NodeLeafs),
	concat(Leafs, NodeLeafs, NewLeafs),
	nodes_leafs(Nodes, NewLeafs, Results).

% enumerate all subtrees from a tree
subtree(Tree, Tree).
subtree([[Child, _], _], Tree) :- subtree(Child, Tree).
subtree([[_, Child], _], Tree) :- subtree(Child, Tree).
subtree(&(List), Subtree) :- element(Tree, List), subtree(Tree, Subtree).

% enumerate all subtrees from a tree (even partial tree => eliminating nodes)
subtree2(Tree, Tree).
subtree2([[Child, _], _], Tree) :- subtree2(Child, Tree).
subtree2([[_, Child], _], Tree) :- subtree2(Child, Tree).
subtree2([[Child1, Child2], Tags], [[Tree1, Tree2], Tags]) :-
	subtree2(Child1, Tree1),
	subtree2(Child2, Tree2),
	(Tree1 \= Child1 ; Tree2 \= Child2).
subtree2(&(List), Subtree) :- element(Tree, List), subtree2(Tree, Subtree).


% build minimum tree for given leafs (eliminating other leafs)
tree_for_leafs([[], Tags], [[], Tags], [[[], Tags]]).
tree_for_leafs([[Child1, _], _], Tree, Leafs) :-
	leaf_list(Child1, Leafs1),
	lists_intersection(Leafs1, Leafs, InterLeafs),
	lists_equivalent(InterLeafs, Leafs),
	tree_for_leafs(Child1, Tree, Leafs), !.
tree_for_leafs([[_, Child2], _], Tree, Leafs) :-
	leaf_list(Child2, Leafs2),
	lists_intersection(Leafs2, Leafs, InterLeafs),
	lists_equivalent(InterLeafs, Leafs),
	tree_for_leafs(Child2, Tree, Leafs), !.
tree_for_leafs([[Child1, Child2], Tags], [[Tree1, Tree2], Tags], Leafs) :-
	leaf_list(Child1, Leafs1),
	leaf_list(Child2, Leafs2),
	concat(Leafs1, Leafs2, AllLeafs),
	lists_intersection(AllLeafs, Leafs, InterLeafs),
	lists_equivalent(InterLeafs, Leafs),
	lists_intersection(Leafs1, Leafs, SLeafs1),
	tree_for_leafs(Child1, Tree1, SLeafs1),
	lists_intersection(Leafs2, Leafs, SLeafs2),
	tree_for_leafs(Child2, Tree2, SLeafs2).
	
% ### tree extractors to get biggest_trees or other predicates working with many kind of lists: "tree" list, "term with tree" list, ...
self_tree_extractor(Tree, Tree).
couple_tree_extractor(Couple, Tree) :- Couple =.. [_, Tree, _].

	
% biggest trees from a list (including others)
biggest_trees([], Result, Result, _).
biggest_trees([Term|Trees], Others, Result, TreeExtractor) :-
	Extract =.. [TreeExtractor, Term, Tree],
	Extract,
	element(OtherTerm, Others),
	ExtractOther =.. [TreeExtractor, OtherTerm, Other],
	ExtractOther,
	match(_, Tree, @(Other)), !,
	without(OtherTerm, Others, NewOthers), 
	biggest_trees(Trees, [Term|NewOthers], Result, TreeExtractor).
biggest_trees([Term|Trees], Others, Result, TreeExtractor) :-
	Extract =.. [TreeExtractor, Term, Tree],
	Extract,
	element(OtherTerm, Others),
	ExtractOther =.. [TreeExtractor, OtherTerm, Other],
	ExtractOther,
	match(_, Other, @(Tree)), !,
	biggest_trees(Trees, Others, Result, TreeExtractor).
biggest_trees([Term|Trees], Others, Result, TreeExtractor) :-
	biggest_trees(Trees, [Term|Others], Result, TreeExtractor).
	
% matching a tree T with a pattern P gives a result tree R
% matching if P is a simple tree structure => R are all trees unifiable with P
% matching if P is a @(Tree) => R are all trees which have Tree as subtree
% matching if P is a $(Tree) => R are all trees, even partial (without some nodes), which have Tree as subtree
% matching if P is a &(ListOfTrees) => R = T if each node element of ListOfTrees is a subtree of T, and all leafs of T are covered
% (if T = &(List), each element of List must match an element in ListOfTrees, and all leafs of T are covered)
% matching if P is a #(Leafs) => R = T if leafs set of T = Leafs  


% match leafs
match_leafs([], []).
match_leafs([Leaf1|Leafs1], Leafs2) :-
	element(Leaf2, Leafs2),
	match(_, Leaf1, Leaf2), !,
	without(Leaf2, Leafs2, LLeafs2),
	match_leafs(Leafs1, LLeafs2).
	

% tree match if all nodes are subtree and leafs list are the same
mixed_match(T, &(Nodes)) :-
	T \= &(_),
	nonvar(Nodes),
	nodes_leafs(Nodes, [], NodesLeafs),
	leaf_list(T, LeafsList),
	match_leafs(LeafsList, NodesLeafs),
	for_each(Nodes, [match_in, T]).
	
% match in
match_in(Tree, Subtree) :- match(_, Tree, @(Subtree)).

% tags matches if same or pattern tags included in given tags
match_tags(Tags, Tags) :- !.
match_tags(_, []) :- !.
match_tags(Tags, [TagTerm|PatternTags]) :-
	element(TagTerm, Tags),
	match_tags(Tags, PatternTags).

% some elements of nodes list match
match_each(&([]), _, Result, Result).
match_each(&([Node1|Nodes1]), &(Nodes2), Matchings, Result) :-
	element(Node2, Nodes2),
	match_in(Node2, Node1), !,
	match_each(&(Nodes1), &(Nodes2), [Node1|Matchings], Result).
match_each(&([Node1|Nodes1]), &(Nodes2), Matchings, Result) :-
	match_each(&(Nodes1), &(Nodes2), Matchings, Result).

% variables in pattern
match(T, T, P) :- var(P), P = T, !.
match([Children, Tags], [Children, Tags], [ChildrenVar, TagsVar]) :-
	var(ChildrenVar),
	var(TagsVar),
	ChildrenVar = Children,
	TagsVar = Tags,
	!.
match([Children, Tags], [Children, Tags], [ChildrenVar, PTags]) :-
	var(ChildrenVar),
	ChildrenVar = Children,
	match_tags(Tags, PTags),
	!.

% tree match if result is leaf list
match(T, T, #(Leafs)) :- leaf_list(T, Leafs1), match_leafs(Leafs, Leafs1).
match(T, T, $(#(Leafs))) :- var(Leafs), leaf_list(T, Leafs), !.
match(T1, T, $(#(Leafs))) :- tree_for_leafs(T, T1, Leafs).

	
% matching for nodes list 
match(T, T, &(Nodes)) :- mixed_match(T, &(Nodes)).
match(&(Nodes), &(Nodes), T) :- mixed_match(T, &(Nodes)).
match(&(Nodes1), &(Nodes1), &(Nodes2)) :-
	nodes_leafs(Nodes1, [], NodesLeafs1),
	nodes_leafs(Nodes2, [], NodesLeafs2),
	match_leafs(NodesLeafs1, NodesLeafs2),
	match_each(&(Nodes1), &(Nodes2), [], N1N2),
	match_each(&(Nodes2), &(Nodes1), [], N2N1), !,
	concat(N1N2, N2N1, AllMatching),
	nodes_leafs(AllMatching, [], AllLeafs),
	only_once(AllLeafs, [], AllLeafsOnce),
	match_leafs(NodesLeafs1, AllLeafsOnce).
	
	

% tree match if identical or if a subtree match
match(S, T, @(Tree)) :-	subtree(T, S), match(_, S, Tree).
	
% tree match if identical or if a subtree (event partial) match producing any result
match(S, T, $(Tree)) :-	Tree \= #(_), subtree2(T, S), match(_, S, Tree).


% leaf matches leaf if tags match
match([[], Tags], [[], Tags], [[], PTags]) :- match_tags(Tags, PTags), !.

% node matches node if children match (in one order or the other)
match([[RChild1, RChild2], Tags], [[Child1, Child2], Tags], [[PChild1, PChild2], PTags]) :-
	match_tags(Tags, PTags),
	match(RChild1, Child1, PChild1),
	match(RChild2, Child2, PChild2).
match([[RChild1, RChild2], Tags], [[Child1, Child2], Tags], [[PChild1, PChild2], PTags]) :-
	match_tags(Tags, PTags),
	match(RChild1, Child1, PChild2),
	match(RChild2, Child2, PChild1).

% matching subtree between two trees
matching_subtree(Tree1, Tree2, Subtree, AcceptLeafs, ResTree) :-
	subtree(Tree1, Subtree),
	(AcceptLeafs = 'true' ; Subtree \= [[], _]),
	one_success(pattern_matching(Tree2, @(Subtree), [], ResTree)).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      Convert to NHX format
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% to NHX
domain_to_nhx(Domain, NHX) :-
	Domain =.. [DomainLabel, Start, End],
	number2atom(Start, StartAtom),
	number2atom(End, EndAtom),
	atom_concat('>', StartAtom, Str1),
	atom_concat(Str1, '>', Str2),
	atom_concat(Str2, EndAtom, Str3),
	atom_concat(Str3, '>', Str4),
	atom_concat(Str4, '0.0', Str5),
	atom_concat(Str5, '>', Str6),
	atom_concat(Str6, DomainLabel, NHX).
	
ds_to_nhx([], Result, Result).
ds_to_nhx([Domain|Left], Init, Result) :-
	domain_to_nhx(Domain, NHX),
	atom_concat(Init, NHX, NewInit),
	ds_to_nhx(Left, NewInit, Result).
ds_nhx(Domains, Length, Result) :- ds_to_nhx(Domains, Length, Result).

to_nhx_duplication(true, 'Y').
to_nhx_duplication(false, 'N').
to_nhx([[], Tags], String) :-
	element(id(ID), Tags),
	element(name(Name), Tags),
	element(species(Species), Tags),
	element(taxeid(Taxeid), Tags),
	element(length(Length), Tags),
	atom_concat(Name, ':', Str1),
	atom_concat(Str1, Length, Str2),
	atom_concat(Str2, '[&&NHX:S=', Str3),
	atom_concat(Str3, Species, Str4),
	atom_concat(Str4, ':T=', Str5),
	atom_concat(Str5, Taxeid, Str6),
	atom_concat(Str6, ':ID=', Str7),
	atom_concat(Str7, ID, Str8),
	
	(element(ds_plength(DSL), Tags) ->
		(
			element(ds(DS), Tags),
			ds_nhx(DS, DSL, DSTag),
			atom_concat(Str8, ':DS=', Str9),
			atom_concat(Str9, DSTag, Str10)
		)
		;
		(
			Str10 = Str8
		)
	),	
	atom_concat(Str10, ']', String).
	
to_nhx([[Child1, Child2], Tags], String) :-
	to_nhx(Child1, SChild1),
	to_nhx(Child2, SChild2),
	element(id(ID), Tags),
	element(name(Name), Tags),
	element(length(Length), Tags),
	element(duplication(Duplication), Tags),
	to_nhx_duplication(Duplication, Dup),
	atom_concat('(', SChild1, Str1),
	atom_concat(Str1, ',', Str2),
	atom_concat(Str2, SChild2, Str3),
	atom_concat(Str3, ')', Str4),
	atom_concat(Str4, Name, Str44),
	atom_concat(Str44, ':', Str5),
	atom_concat(Str5, Length, Str6),
	atom_concat(Str6, '[&&NHX:D=', Str7),
	atom_concat(Str7, Dup, Str8),
	atom_concat(Str8, ':ID=', Str9),
	atom_concat(Str9, ID, Str10),
	atom_concat(Str10, ']', String).
	
to_nhx(&([Tree|Trees]), String) :-
	Trees \= [],
	to_nhx(Tree, SChild1),
	to_nhx(&(Trees), SChild2),
	Length = 0,
	Name = '_',
	to_nhx_duplication(false, Dup),
	atom_concat('(', SChild1, Str1),
	atom_concat(Str1, ',', Str2),
	atom_concat(Str2, SChild2, Str3),
	atom_concat(Str3, ')', Str4),
	atom_concat(Str4, Name, Str44),
	atom_concat(Str44, ':', Str5),
	atom_concat(Str5, Length, Str6),
	atom_concat(Str6, '[&&NHX:D=', Str7),
	atom_concat(Str7, Dup, Str8),
	atom_concat(Str8, ']', String), !.
	
to_nhx(&(Tree), String) :-	to_nhx(Tree, String).

nhx(Tree, String) :-
	to_nhx(Tree, Str),
	atom_concat(Str, ';', String).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	