/* PhyloPattern
 * Copyright (C) 2005-2009 Philippe Gouret
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA. The text ol license can be also found 
 * at http://www.gnu.org/copyleft/gpl.html
 */

//------------------------------------------------------------------------------
// nhx parser caller
// Philippe Gouret
//------------------------------------------------------------------------------

// prolog buildins package
package gnu.prolog.vm.buildins.PhyloPattern;

// import classes
import gnu.prolog.term.*;
import gnu.prolog.vm.*;
import gnu.prolog.io.*;
import java.io.*;
import java.util.*;
import org.forester.phylogeny.*;
import org.forester.io.parsers.nhx.*; 
import org.forester.phylogeny.data.*;


// project classes

/** prolog code
 */
public class Predicate_nhx_as_term implements PrologCode, TermConstants
{
	// opened bracket
	static private final char OPENED_BRACKET = '[';
	// closed bracket
	static private final char CLOSED_BRACKET = ']';

	// opened brace
	static private final char OPENED_BRACE = '{';
	// closed brace
	static private final char CLOSED_BRACE = '}';

	// dot
	static private final char DOT = '.';
	// comma
	static private final char COMMA = ',';
	// quote
	static private final char QUOTE = '\'';
	
	// automatic ids index
	private int automaticIdIndex;

	/**
	 *  generated term associated with node name
	 *  @param name to parse
	 *  @return term built from name
	 */
	private Term parsedName(String name)
	{
		// replace special chars: "{aaa.bbb.1}" by "['aaa','bbb',1]"
		// ### nhx format doesn't accept these characters ('{' and '.')
		// ### start at end cause '.' char can be found in original name
		boolean opened = false;
		boolean closed = false;
		int nbDots = 0;
		String newName = "";
		for (int i=name.length()-1; i>=0; i--)
		{
			// '}' found
			if (!opened && !closed && (nbDots == 0) && (name.charAt(i) == CLOSED_BRACE))
			{
				newName = CLOSED_BRACKET + newName;
				closed = true;
			}
			// '{'
			else if (!opened && closed && (nbDots == 2) && (name.charAt(i) == OPENED_BRACE))
			{
				newName = QUOTE + newName;
				newName = OPENED_BRACKET + newName;
				opened = true;
			}
			// '.'
			else if (!opened && closed && (nbDots < 2) && (name.charAt(i) == DOT))
			{
				if (nbDots == 0)
				{
					newName = COMMA + newName;
					newName = QUOTE + newName;
				}
				if (nbDots == 1)
				{
					newName = QUOTE + newName;
					newName = COMMA + newName;
					newName = QUOTE + newName;
				}
				nbDots++;
			}
			else newName = name.charAt(i) + newName;
		}

		// bad formet
		if (!opened || !closed || (nbDots != 2)) return AtomTerm.get(name);

		try
		{
			// Prolog parse possible
			return TermReader.stringToTerm(newName);
		}
		catch (Exception e)
		{
			StringWriter trace = new StringWriter();
			e.printStackTrace(new PrintWriter(trace));
			// ### bad management
			return null;
		}
	}

	/**
	 *  build a term containing all nhx hierarchy
	 *  @param nhx tree node element
	 *  @return built term
	 */
	private Term generateTerm(PhylogenyNode node)
	{
		// get root fields
		String name = node.getName();
		if (name.equalsIgnoreCase("")) name = "_";
		double length = node.getDistanceToParent();
		NodeData nodeData = node.getNodeData();
		BranchData branchData = node.getBranchData();
		Taxonomy taxonomy = nodeData.getTaxonomy(); 
		
		// bootstraps
		int bootstrap = branchData.isHasConfidences() ? ((int) branchData.getConfidence(0).getValue() * 100) : -1;
		
		// species
		String species = (nodeData.isHasTaxonomy()) ? taxonomy.getTaxonomyCode() : "?";
		if (species.equalsIgnoreCase("")) species = "?";
		int taxeid = -1;
		try
		{
			taxeid = (nodeData.isHasTaxonomy()) ? Integer.parseInt(nodeData.getTaxonomy().getIdentifier().getValue()) : -1;
		}
		catch (Exception e) {};

		// id
		String id = null;
		NodeData data = node.getNodeData(); 
		if (data != null)
		{
			Identifier identifier = data.getNodeIdentifier();
			if (identifier != null)
				id = identifier.getValue();
			else
				id = Integer.toString(automaticIdIndex++);
		}
		else id = Integer.toString(automaticIdIndex++);

		// duplication
		String duplication = Boolean.toString(node.isDuplication()); 
		
		// sequence if any
		Term domainsList = null;
		DomainArchitecture architecture = null;
		if (nodeData.isHasSequence())
		{
			// domains architecture if any
			List<Term> domains = new Vector<Term>();
			architecture = nodeData.getSequence().getDomainArchitecture();
			if (architecture != null)
			{
				// for each domain
				int nbDomains = architecture.getDomains().keySet().size();
				for (int i=0; i<nbDomains; i++)
				{
					// one
					ProteinDomain domain = architecture.getDomain(i);
					
					// domain term
					CompoundTerm domainTerm =
						new CompoundTerm(domain.getName(), new Term[]{new IntegerTerm(domain.getFrom()),new IntegerTerm(domain.getTo())});
					domains.add(domainTerm);
				}

				// domains list
				domainsList = CompoundTerm.getList(domains);
			}
		}
		
		// get child1 term if any
		Term child1Term = null;
		if (node.isInternal())
		{
			PhylogenyNode child1 = node.getChildNode1();
			if (child1 != null) child1Term = generateTerm(child1);
		}
		
		// get child2 term if any
		Term child2Term = null;
		if (node.isInternal())
		{
			PhylogenyNode child2 = node.getChildNode2();
			if (child2 != null) child2Term = generateTerm(child2);
		}
		
		// build term for node
		Term nameTerm = parsedName(name);
		if (nameTerm == null) return null;

		// children
		List<Term> childList = new Vector<Term>();
		if (child1Term != null) childList.add(child1Term);
		if (child2Term != null) childList.add(child2Term);
		Term childrenList = CompoundTerm.getList(childList);

		// attributes
		List<Term> attrList = new Vector<Term>();
		attrList.add(new CompoundTerm("name", new Term[]{nameTerm}));
		attrList.add(new CompoundTerm("id", new Term[]{AtomTerm.get(id)}));
		attrList.add(new CompoundTerm("length", new Term[]{AtomTerm.get(Double.toString(length))}));
		attrList.add(new CompoundTerm("bootstrap", new Term[]{AtomTerm.get(Integer.toString(bootstrap))}));
		attrList.add(new CompoundTerm("species", new Term[]{AtomTerm.get(species)}));
		attrList.add(new CompoundTerm("taxeid", new Term[]{AtomTerm.get(Integer.toString(taxeid))}));
		if (architecture != null) attrList.add(new CompoundTerm("ds_plength", new Term[]{AtomTerm.get(Integer.toString(architecture.getTotalLength()))}));
		attrList.add(new CompoundTerm("duplication", new Term[]{AtomTerm.get(duplication)}));
		if (domainsList != null) attrList.add(new CompoundTerm("ds", new Term[]{domainsList}));
		Term attributes_list = CompoundTerm.getList(attrList);

		// tree
		List<Term> tree = new Vector<Term>();
		tree.add(childrenList);
		tree.add(attributes_list);
		Term result = CompoundTerm.getList(tree);
		return result;
	}

	/** this method is used for execution of code
	 * @param interpreter interpreter in which context code is executed
	 * @param backtrackMode true if predicate is called on backtracking and false otherwise
	 * @param args arguments of code
	 * @return either SUCCESS, SUCCESS_LAST, or FAIL.
	 */
	public int execute(Interpreter interpreter, boolean backtrackMode, gnu.prolog.term.Term args[])
	throws PrologException
	{
		// get parameters
		Term inTerm = args[0];
		Term outTerm = args[1];

		// in
		if (!(inTerm instanceof AtomTerm)) PrologException.typeError(atomAtom, inTerm);
		String in = ((AtomTerm) inTerm).value;
		if (in == null) return FAIL;
		if (in.equals("")) return FAIL;

		// out
		if (!(outTerm instanceof VariableTerm)) PrologException.typeError(variableAtom, outTerm);
		VariableTerm out = (VariableTerm) outTerm;

		try
		{
			// build tree from nhx
			NHXParser parser = new NHXParser();
			parser.setSource(in);
			Phylogeny tree = parser.parse()[0];

			// generate prolog term from nhx tree
			// affect out variable
			automaticIdIndex = 0;
			interpreter.addVariableUndo(out);
			Term res = generateTerm(tree.getRoot());
			if (res != null) out.value = res;
		}
		catch (Exception e)
		{
			// trace
			e.printStackTrace();
			
			// failed
			return FAIL;
		}

		// success
		return SUCCESS_LAST;
	}

	/** this method is called when code is installed to the environment
	 * code can be installed only for one environment.
	 * @param environment environemnt to install the predicate
	 */
	public void install(Environment env)
	{

	}

	/** this method is called when code is uninstalled from the environment
	 * @param environment environemnt to install the predicate
	 */
	public void uninstall(Environment env)
	{
	}

}

